/*
 * Copyright 1999, Alexander Feldman <alex@varna.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Alexander Feldman nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ALEXANDER FELDMAN AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL ALEXANDER FELDMAN OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <syslog.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <glob.h>

#include "config.h"
#include "globals.h"
#include "database.hpp"
#include "obscure.h"
#include "server.hpp"
#include "sockio.h"
#include "salt.h"
#include "conf.hpp"
#include "group.hpp"
#include "tokenized_string.hpp"
#include "trim.h"

#ifdef HAVE_LIBCRYPT
	extern "C" char *crypt(const char *, const char *);
#endif

extern CConfigurationFile *pConfigFile;

void server(int sockfd)
{
	char szLogin[32];
	char *pszPassword = NULL;
	
	write_byte(sockfd, WELCOME);
	if (-1 == read_string(sockfd, szLogin, sizeof(szLogin)))
		write_byte(sockfd, TIMEOUT);
	if (0 != strcmp(szLogin, "@key@")) {
		if (NULL == (pszPassword = read_password(sockfd)))
			write_byte(sockfd, TIMEOUT);
		if (false == check_login(szLogin, pszPassword)) {
			write_byte(sockfd, LOGIN_FAILED);
			return;
		}
	}
	write_byte(sockfd, GO_AHEAD);
	
	unsigned char cAction = '\0';
	if (true != read_byte(sockfd, &cAction))
		write_byte(sockfd, TIMEOUT);
	
	switch (cAction) {
		case ACTION_GETKEY:
			action_getkey(sockfd);
		break;
		case ACTION_NOTHING:
			syslog(LOG_USER | LOG_INFO, "user %s did nothing", szLogin);
			write_byte(sockfd, GOOD_BYE);
		break;
		case ACTION_CHANGEGROUPPASSWORD:
			action_change_group_password(sockfd, szLogin, pszPassword);
		break;
		case ACTION_CHANGEPASSWORD:
			action_change_password(sockfd, szLogin, pszPassword);
		break;
		case ACTION_ADDGROUP:
			action_add_group(sockfd, szLogin);
		break;
		case ACTION_CHFN:
			action_chfn(sockfd, szLogin);
		break;
		case ACTION_ADDUSER:
			action_add_user(sockfd, szLogin);
		break;
		case ACTION_DELETEGROUP:
			action_delete_group(sockfd, szLogin);
		break;
		case ACTION_DELETEUSER:
			action_delete_user(sockfd, szLogin);
		break;
		case ACTION_GETGROUP:
			action_get_group(sockfd, szLogin);
		break;
		default:
			write_byte(sockfd, INVALID_ACTION);
		break;
	}
	if (NULL != pszPassword)
		free(pszPassword);
	
//	write_byte(sockfd, GOOD_BYE);
}

bool check_gecos_field(char *pszString) 
{
	for (int q = 0; '\0' != pszString[q]; q++)
		if (pszString[q] != ' ' &&
			 (!isalnum(pszString[q]) ||
			  ',' == pszString[q] ||
			  ';' == pszString[q] ||
			  '=' == pszString[q]))
			return false;
	return true;
}

bool allow_change(char *pszLogin, char *pszTarget)
{
	char *pszAllowed;
	char *pszPowered;
	char *pszSheep;
	bool fgAllowed = false;
	bool fgPowered = false;
	bool fgSheep = false;
	
	CPasswdDatabase cPasswdDatabase(true);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");		
		return false;
	}
	CGroupDatabase cGroupDatabase(true);
	if (NO_ERROR != cGroupDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading group file");		
		return false;
	}
	
	CTokenizedString cAllowedUsers(pConfigFile->GetString("Permissions", "users_allow", ""), ";,");
	CTokenizedString cPoweredUsers(pConfigFile->GetString("Permissions", "master", ""), ";,");
	
	gid_t iLoginGroup;
	gid_t iTargetGroup;
	
	if (false == cPasswdDatabase.GetGroup(pszLogin, &iLoginGroup)) {
		syslog(LOG_USER | LOG_INFO, "invalid group for user %s", pszLogin);
		return false;
	}
	if (false == cPasswdDatabase.GetGroup(pszTarget, &iTargetGroup)) {
		syslog(LOG_USER | LOG_INFO, "invalid group for user %s", pszTarget);
		return false;
	}
	
	char *pszLoginGroup = cGroupDatabase.GetName(iLoginGroup);
	char *pszTargetGroup = cGroupDatabase.GetName(iTargetGroup);
	
	if (NULL == pszLoginGroup) {
		syslog(LOG_USER | LOG_INFO, "invalid group for user %s", pszLogin);
		return false;
	}
	if (NULL == pszTargetGroup) {
		syslog(LOG_USER | LOG_INFO, "invalid group for user %s", pszTarget);
		return false;
	}
	if ((pszAllowed = cAllowedUsers.GetFirstString()))
		do {
			pszAllowed = TrimBoth(pszAllowed, " \t\x0a\x0d");
			if ('@' == pszAllowed[0]) {
			  	if (0 == strcmp(pszTargetGroup, pszAllowed + 1)) {
					fgAllowed = true;
					break;
				}
			} else {
			   if (0 == strcmp(pszTarget, pszAllowed)) {
					fgAllowed = true;
					break;
				}
			}
		} while ((pszAllowed = cAllowedUsers.GetNextString()));

	if (false == fgAllowed) {
		syslog(LOG_USER | LOG_INFO, "unauthorized user %s", pszLogin);
		return false;
	}
	if (0 == strcmp(pszLogin, pszTarget))
		return true;

	if ((pszPowered = cPoweredUsers.GetFirstString()))
		do {
			pszPowered = TrimBoth(pszPowered, " \t\x0a\x0d");
			if (NULL == (pszSheep = strchr(pszPowered, ':'))) {
				syslog(LOG_USER | LOG_INFO, "configuration file error, skipping power users");
				continue;
			}
			*pszSheep = '\0';
			pszSheep += 1;
			if ('@' == pszPowered[0]) {
				if (0 == strcmp(pszPowered + 1, pszLoginGroup))
					fgPowered = true;
			} else {
				if (0 == strcmp(pszPowered, pszLogin))
					fgPowered = true;
			}
			if (fgPowered && ('@' == pszSheep[0])) {
				if (0 == strcmp(pszSheep + 1, pszTargetGroup)) {
					fgSheep = true;
					break;
				}
			} else {
				if (0 == strcmp(pszSheep, pszTarget)) {
					fgSheep = true;
					break;
				}
			}
		} while ((pszPowered = cPoweredUsers.GetNextString()));
	
	if (true == fgSheep)
		return true;
	
	syslog(LOG_USER | LOG_INFO, "%s tried to change the password of %s without rights", pszLogin, pszTarget);
	
	return false;
}

bool allow_change_group(char *pszLogin, char *pszTarget)
{
	char *pszPowered;
	char *pszSheep;
	bool fgPowered = false;
	bool fgSheep = false;
	
	CPasswdDatabase cPasswdDatabase(true);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");		
		return false;
	}
	CGroupDatabase cGroupDatabase(true);
	if (NO_ERROR != cGroupDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading group file");		
		return false;
	}
	
	gid_t iLoginGroup;
	
	if (false == cPasswdDatabase.GetGroup(pszLogin, &iLoginGroup)) {
		syslog(LOG_USER | LOG_INFO, "invalid group for user %s", pszLogin);
		return false;
	}
	
	char *pszLoginGroup = cGroupDatabase.GetName(iLoginGroup);
	
	if (NULL == pszLoginGroup) {
		syslog(LOG_USER | LOG_INFO, "invalid group for user %s", pszLogin);
		return false;
	}
	
	CTokenizedString cPoweredUsers(pConfigFile->GetString("Permissions", "master", ""), ";,");
	
	if ((pszPowered = cPoweredUsers.GetFirstString()))
		do {
			pszPowered = TrimBoth(pszPowered, " \t\x0a\x0d");
			if (NULL == (pszSheep = strchr(pszPowered, ':'))) {
				syslog(LOG_USER | LOG_INFO, "configuration file error, skipping power users");
				continue;
			}
			*pszSheep = '\0';
			pszSheep += 1;
			if ('@' == pszPowered[0]) {
				if (0 == strcmp(pszPowered + 1, pszLoginGroup))
					fgPowered = true;
			} else {
				if (0 == strcmp(pszPowered, pszLogin))
					fgPowered = true;
			}
			if (fgPowered && ('@' == pszSheep[0]))
				if (0 == strcmp(pszSheep + 1, pszTarget)) {
					fgSheep = true;
					break;
				}
		} while ((pszPowered = cPoweredUsers.GetNextString()));
	
	if (true == fgSheep)
		return true;
	
#ifdef SHADOWGRP
	uid_t i = getuid();
	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to change passwords; setuid error: %s", strerror(errno));
		return false;
	}												// With shadow enabled we need to be root to
													// open the file
	CGshadowDatabase cGshadowDatabase(true);
	if (NO_ERROR != cGshadowDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading gshadow file");
		return false;
	}
	if (true == cGshadowDatabase.IsAdministrator(pszTarget, pszLogin))
		return true;
	setuid(i);
#endif	
	
	syslog(LOG_USER | LOG_INFO, "%s tried to change the group password of %s without rights", pszLogin, pszTarget);
	
	return false;
}

bool change_password(char *pszLogin,
							char *pszNewPassword,
							int iMin,
							int iMax,
							int iWarn,
							int iInact)
{
	char *pszCryptedPassword = crypt(pszNewPassword, crypt_make_salt());
	
	memset(pszNewPassword, 0, strlen(pszNewPassword));		// Clear password

	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to change passwords setuid error: %s", strerror(errno));
		return false;
	}
	
#ifdef SHADOWPWD
	CShadowDatabase cShadowDatabase(false);
	if (NO_ERROR != cShadowDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading shadow file");
		return false;
	}
	if (false == cShadowDatabase.ChangePassword(pszLogin,
															  pszCryptedPassword,
															  iMin,
															  iMax,
															  iWarn,
															  iInact)) {
		syslog(LOG_USER | LOG_INFO, "error modifying shadow file");
		return false;
	}
#else
	CPasswdDatabase cPasswdDatabase(false);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		return false;
	}
	if (false == cPasswdDatabase.ChangePassword(pszLogin,
															  pszCryptedPassword,
															  iMin,
															  iMax,
															  iWarn,
															  iInact)) {
		syslog(LOG_USER | LOG_INFO, "error modifying passwd file");
		return false;
	}
#endif
	
	return true;
}

bool change_gecos(char *pszLogin,
						char *pszFullName,
						char *pszRoomNumber,
						char *pszWorkPhone,
						char *pszHomePhone,
						char *pszOther)
{
	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to change passwords setuid error: %s", strerror(errno));
		return false;
	}

	CPasswdDatabase cPasswdDatabase(false);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		return false;
	}
	char *pszOldGecos = cPasswdDatabase.GetGecos(pszLogin);
	if (NULL == pszOldGecos) {
		syslog(LOG_USER | LOG_INFO, "error getting old gecos field");
		return false;
	}
	CTokenizedString cOldGecos(pszOldGecos, ",");
	if (*pszFullName == '\0' && cOldGecos.GetString(0))
		pszFullName = cOldGecos.GetString(0);
	if (*pszRoomNumber == '\0' && cOldGecos.GetString(1))
		pszRoomNumber = cOldGecos.GetString(1);
	if (*pszWorkPhone == '\0' && cOldGecos.GetString(2))
		pszWorkPhone = cOldGecos.GetString(2);
	if (*pszHomePhone == '\0' && cOldGecos.GetString(3))
		pszHomePhone = cOldGecos.GetString(3);
	if (*pszOther == '\0' && cOldGecos.GetString(4))
		pszOther = cOldGecos.GetString(4);
	char *pszNewGecos = new char[1024];
	if (NULL == pszNewGecos) {
		syslog(LOG_USER | LOG_INFO, "memory allocation error");
		return false;
	}
	strcpy(pszNewGecos, pszFullName);
	strcat(pszNewGecos, ",");
	strcat(pszNewGecos, pszRoomNumber);
	strcat(pszNewGecos, ",");
	strcat(pszNewGecos, pszWorkPhone);
	strcat(pszNewGecos, ",");
	strcat(pszNewGecos, pszHomePhone);
	if ('\0' != *pszOther) {
		strcat(pszNewGecos, ",");
		strcat(pszNewGecos, pszOther);
	}
	if (false == cPasswdDatabase.ChangeGecos(pszLogin, pszNewGecos)) {
		syslog(LOG_USER | LOG_INFO, "error modifying passwd file");
		delete pszNewGecos;
		return false;
	}
	delete pszNewGecos;
	
	return true;
}

bool change_password_group(char *pszLogin, char *pszNewPassword)
{
	char *pszCryptedPassword = crypt(pszNewPassword, crypt_make_salt());
	
	memset(pszNewPassword, 0, strlen(pszNewPassword));		// Clear password

	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to change passwords setuid error: %s", strerror(errno));
		return false;
	}
	
#ifdef SHADOWGRP
	CGshadowDatabase cGshadowDatabase(false);
	if (NO_ERROR != cGshadowDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading gshadow file");
		return false;
	}
	if (false == cGshadowDatabase.ChangePassword(pszLogin, pszCryptedPassword, 0, 0, 0, 0)) {
		syslog(LOG_USER | LOG_INFO, "error modifying gshadow file");
		return false;
	}
#else
	CGroupDatabase cGroupDatabase(false);
	if (NO_ERROR != cGroupDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading group file");
		return false;
	}
	if (false == cGroupDatabase.ChangePassword(pszLogin, pszCryptedPassword, 0, 0, 0, 0)) {
		syslog(LOG_USER | LOG_INFO, "error modifying group file");
		return false;
	}
#endif
	
	return true;
}

bool check_login(char *pszLogin, char *pszPassword)
{
// Check to see if there is a valid login name
	CPasswdDatabase cPasswdDatabase(true);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		return false;
	}
	if (false == cPasswdDatabase.FindUser(pszLogin)) {	
		syslog(LOG_USER | LOG_INFO, "invalid user: %s", pszLogin);
		return false;
	}
#ifdef SHADOWPWD
	uid_t i = getuid();
	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to change passwords setuid error: %s", strerror(errno));
		return false;
	}												// With shadow enabled we need to be root to
													// open the file
	CShadowDatabase cShadowDatabase(true);
	if (NO_ERROR != cShadowDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading shadow file");
		return false;
	}
	if (false == cShadowDatabase.FindUser(pszLogin)) {
		syslog(LOG_USER | LOG_INFO, "invalid user: %s", pszLogin);
		return false;
	}
	char *pszUserPassword = cShadowDatabase.GetPassword(pszLogin);
	if ((NULL == pszUserPassword) ||
		 (0 != strcmp(pszUserPassword, crypt(pszPassword, pszUserPassword)))) {
		syslog(LOG_USER | LOG_INFO, "invalid password for user: %s", pszLogin);
		return false;
	}
#else
	char *pszUserPassword = cPasswdDatabase.GetPassword(pszLogin);
	if ((NULL == pszUserPassword) || 
		 (0 != strcmp(pszUserPassword, crypt(pszPassword, pszUserPassword)))) {
		syslog(LOG_USER | LOG_INFO, "invalid password for user: %s", pszLogin);
		return false;
	}
#endif
#ifdef SHADOWPWD
	setuid(i);									// Return to the original uid, paranoid, huh?
#endif	
	return true;
}

bool check_target(char *pszTarget)
{
// Check to see if there is a valid login name
	CPasswdDatabase cPasswdDatabase(true);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		return false;
	}
	if (false == cPasswdDatabase.FindUser(pszTarget)) {	
//		syslog(LOG_USER | LOG_INFO, "invalid user: %s", pszTarget);
		return false;
	}
#ifdef SHADOWPWD
	uid_t i = getuid();
	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to change passwords setuid error: %s", strerror(errno));
		return false;
	}												// With shadow enabled we need to be root to
													// open the file
	CShadowDatabase cShadowDatabase(true);
	if (NO_ERROR != cShadowDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading shadow file");
		return false;
	}
	if (false == cShadowDatabase.FindUser(pszTarget)) {
//		syslog(LOG_USER | LOG_INFO, "invalid user: %s", pszTarget);
		return false;
	}
	setuid(i);
#endif	
	return true;
}

bool check_target_group(char *pszTarget)
{
// Check to see if there is a valid login name
	CGroupDatabase cGroupDatabase(true);
	if (NO_ERROR != cGroupDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading group file");
		return false;
	}
	if (false == cGroupDatabase.FindUser(pszTarget)) {	
//		syslog(LOG_USER | LOG_INFO, "invalid group: %s", pszTarget);
		return false;
	}
#ifdef SHADOWGRP
	uid_t i = getuid();
	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to change passwords setuid error: %s", strerror(errno));
		return false;
	}												// With shadow enabled we need to be root to
													// open the file
	CGshadowDatabase cGshadowDatabase(true);
	if (NO_ERROR != cGshadowDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading gshadow file");
		return false;
	}
	if (false == cGshadowDatabase.FindUser(pszTarget)) {
//		syslog(LOG_USER | LOG_INFO, "invalid user: %s", pszTarget);
		return false;
	}
	setuid(i);
#endif	
	return true;
}

bool delete_user(char *pszTarget)
{
	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to delete user; setuid error: %s", strerror(errno));
		return false;
	}
	
	CPasswdDatabase cPasswdDatabase(false);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		return false;
	}
	
	char *pszHome = cPasswdDatabase.GetHome(pszTarget);
	if (NULL == pszHome) {
		syslog(LOG_USER | LOG_INFO, "error getting home directory for user %s", pszTarget);
		return false;
	}
	pszHome = strdup(pszHome);
	if (NULL == pszHome) {
		syslog(LOG_USER | LOG_INFO, "memory allcoation error");
		return false;
	}
	
	if (false == cPasswdDatabase.DeleteEntry(pszTarget)) {
		syslog(LOG_USER | LOG_INFO, "error modifying passwd file");
		delete pszHome;
		return false;
	}
	
#ifdef SHADOWPWD
	CShadowDatabase cShadowDatabase(false);
	if (NO_ERROR != cShadowDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading shadow file");
		delete pszHome;
		return false;
	}
	if (false == cShadowDatabase.DeleteEntry(pszTarget)) {
		syslog(LOG_USER | LOG_INFO, "error modifying shadow file");
		delete pszHome;
		return false;
	}
#endif

	char *pszTrash = pConfigFile->GetString("Delete", "trash", "/tmp");
	if (pConfigFile->GetBoolean("Delete", "move", false)) {
// Move home directory to trash
		char *pszNew = new char[1024];
		if (NULL == pszNew) {
			syslog(LOG_USER | LOG_INFO, "memory allcoation error");
			delete pszHome;
			return false;
		}
		strcpy(pszNew, pszTrash);
		if (NULL != strrchr(pszHome, '/'))
			strcat(pszNew, strrchr(pszHome, '/'));
		else
			strcat(pszNew, pszHome);
		if (0 != rename(pszHome, pszNew)) {
			syslog(LOG_USER | LOG_INFO, "error moving home directory '%s' to trash directory '%s', %s", pszHome, pszNew, strerror(errno));
			delete pszNew;
			delete pszHome;
			return false;
		}
		syslog(LOG_USER | LOG_INFO, "moving home directory '%s' to trash directory '%s'", pszHome, pszNew);
		delete pszNew;		
	}
	if (pConfigFile->GetBoolean("Delete", "remove", false)) {
// Remove home directory
		if (false == remove_tree(pszHome)) {
			syslog(LOG_USER | LOG_INFO, "error removing home directory '%s'", pszHome);
			delete pszHome;
			return false;
		}
		syslog(LOG_USER | LOG_INFO, "removing home directory '%s'", pszHome);
	} 
	delete pszHome;
	
	return true;
}

bool delete_group(char *pszTarget)
{
	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to delete group; setuid error: %s", strerror(errno));
		return false;
	}
	
	CGroupDatabase cGroupDatabase(false);
	if (NO_ERROR != cGroupDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading group file");
		return false;
	}
	if (false == cGroupDatabase.DeleteEntry(pszTarget)) {
		syslog(LOG_USER | LOG_INFO, "error modifying group file");
		return false;
	}

#ifdef SHADOWGRP
	CGshadowDatabase cGshadowDatabase(false);
	if (NO_ERROR != cGshadowDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading gshadow file");
		return false;
	}
	if (false == cGshadowDatabase.DeleteEntry(pszTarget)) {
		syslog(LOG_USER | LOG_INFO, "error modifying gshadow file");
		return false;
	}
#endif
	
	return true;
}

bool allow_add(char *pszLogin)
{
	CPasswdDatabase cPasswdDatabase(true);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		return false;
	}
	CGroupDatabase cGroupDatabase(true);
	if (NO_ERROR != cGroupDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading group file");
		return false;
	}
	
	CTokenizedString cAllowedUsers(pConfigFile->GetString("Add", "master", ""), ",;");
	
	char *pszAllowed;
	char *pszGroup;

	gid_t iLoginGroup;
	if (false == cPasswdDatabase.GetGroup(pszLogin, &iLoginGroup)) {
		syslog(LOG_USER | LOG_INFO, "invalid group for user %s", pszLogin);
		return false;
	}
	pszGroup = cGroupDatabase.GetName(iLoginGroup);
	if (NULL == pszGroup) {
		syslog(LOG_USER | LOG_INFO, "invalid group for user %s", pszLogin);
		return false;
	}
	
	if ((pszAllowed = cAllowedUsers.GetFirstString()))
		do {
			pszAllowed = TrimBoth(pszAllowed, " \t\x0a\x0d");
			if ('@' == pszAllowed[0]) {
			  	if (0 == strcmp(pszGroup, pszAllowed + 1))
					return true;
			} else {
			   if (0 == strcmp(pszLogin, pszAllowed))
					return true;
			}
		} while ((pszAllowed = cAllowedUsers.GetNextString()));
	
	return false;
}

bool add_user(char *pszTarget, char *pszGecos, char *pszDirectory, char *pszShell, int iUserID, int iGroupID)
{
	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to add users; setuid error: %s", strerror(errno));
		return false;
	}
	
	CPasswdDatabase cPasswdDatabase(false);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		return false;
	}

	if (false == cPasswdDatabase.AddEntry(pszTarget, iUserID, iGroupID, pszGecos, pszDirectory, pszShell)) {
		syslog(LOG_USER | LOG_INFO, "error modifying shadow file");
		return false;
	}

#ifdef SHADOWPWD
	CShadowDatabase cShadowDatabase(false);
	if (NO_ERROR != cShadowDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading shadow file");
		return false;
	}
	if (false == cShadowDatabase.AddEntry(pszTarget)) {
		syslog(LOG_USER | LOG_INFO, "error modifying shadow file");
		return false;
	}
#endif

	return true;
}

bool add_group(char *pszTarget, char *pszMembers, char *pszAdministrators)
{
	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to add group; setuid error: %s", strerror(errno));
		return false;
	}
	
	CGroupDatabase cGroupDatabase(false);
	if (NO_ERROR != cGroupDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading group file");
		return false;
	}

	unsigned int iGroupID = cGroupDatabase.GetMaxGroupID() + 1;
	
	if (false == cGroupDatabase.AddEntry(pszTarget, iGroupID, pszMembers)) {
		syslog(LOG_USER | LOG_INFO, "error modifying group file");
		return false;
	}

#ifdef SHADOWGRP
	CGshadowDatabase cGshadowDatabase(false);
	if (NO_ERROR != cGshadowDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading gshadow file");
		return false;
	}
	if (false == cGshadowDatabase.AddEntry(pszTarget, pszAdministrators, pszMembers)) {
		syslog(LOG_USER | LOG_INFO, "error modifying gshadow file");
		return false;
	}
#endif
	
	return true;
}

char *get_group(char *pszTarget)
{
	CPasswdDatabase cPasswdDatabase(false);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		return "";
	}
	CGroupDatabase cGroupDatabase(false);
	if (NO_ERROR != cGroupDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading group file");
		return "";
	}
	gid_t iGroup;
	if (false == cPasswdDatabase.GetGroup(pszTarget, &iGroup))
		return "";
	char *pszResult = cGroupDatabase.GetName(iGroup);
	if ((char *)NULL == pszResult)
		return "";
	
	return pszResult;
}

bool remove_tree(char *pszPath)
{
	bool fgResult = true;
	
	glob_t globbuf;
   
	char szTmp[512];
	strncpy(szTmp, pszPath, sizeof(szTmp));
	if (szTmp[strlen(szTmp) - 1] != '/')
		strncat(szTmp, "/", sizeof(szTmp));
	strncat(szTmp, "*", sizeof(szTmp));
	
	if (0 == glob(szTmp, GLOB_PERIOD | GLOB_MARK | GLOB_NOSORT, NULL, &globbuf)) {
		int i = 0;
		while((globbuf.gl_pathv[i])) {
			if ((globbuf.gl_pathv[i][strlen(globbuf.gl_pathv[i]) - 1] == '/') &&
				 (strstr(globbuf.gl_pathv[i], "/../") == (char *)NULL) &&
				 (strstr(globbuf.gl_pathv[i], "/./") == (char *)NULL)) {
				strncpy(szTmp, globbuf.gl_pathv[i], sizeof(szTmp));
				strncat(szTmp, "*", sizeof(szTmp));
				if (0 != glob(szTmp, GLOB_PERIOD | GLOB_MARK | GLOB_APPEND | GLOB_NOSORT, NULL, &globbuf))
					fgResult = false;
			}
			i += 1;
		}

		i = 0;
		while((globbuf.gl_pathv[i]))
			if (globbuf.gl_pathv[i][strlen(globbuf.gl_pathv[i]) - 1] == '/')
				rmdir(globbuf.gl_pathv[i++]);
			else
				remove(globbuf.gl_pathv[i++]);
		
		rmdir(pszPath);
	
		globfree(&globbuf);
	} else
		return false;
                                                                      
	return fgResult;
}

void action_delete_group(int sockfd, char *pszLogin)
{
	char szTarget[32];
	
	if (-1 == read_string(sockfd, szTarget, sizeof(szTarget))) {
		write_byte(sockfd, TIMEOUT);
		return;
	}
	
	syslog(LOG_USER | LOG_INFO, "user %s tries to delete group %s", pszLogin, szTarget);
	
	if (false == check_target_group(szTarget)) {
		write_byte(sockfd, INVALID_USER);
		return;
	}
	if (false == allow_change_group(pszLogin, szTarget)) {
		write_byte(sockfd, NO_RIGHTS);
		return;
	}
	uid_t i = getuid();
	if (false == delete_group(szTarget)) {
		write_byte(sockfd, ERROR);
		setuid(i);
		return;
	}
	write_byte(sockfd, OK);
	setuid(i);
}

void action_change_group_password(int sockfd, char *pszLogin, char *pszPassword)
{
	char szTarget[32];
	char *pszNewPassword = NULL;
	
	if ((-1 == read_string(sockfd, szTarget, sizeof(szTarget))) ||
		 (NULL == (pszNewPassword = read_password(sockfd)))) {
		write_byte(sockfd, TIMEOUT);
		return;
	}
	
	syslog(LOG_USER | LOG_INFO, "user %s begins changing password of group %s", pszLogin, szTarget);
	if (false == check_target_group(szTarget)) {
		free(pszNewPassword);
		write_byte(sockfd, INVALID_USER);
		return;
	}
	if (false == allow_change_group(pszLogin, szTarget)) {
		write_byte(sockfd, NO_RIGHTS);
		free(pszNewPassword);
		return;
	}
	char *pszObscure = obscure(pszPassword, pszNewPassword, false);
	if (NULL != pszObscure) {
		free(pszNewPassword);
		write_byte(sockfd, WEAK_PASSWORD);
		write_string(sockfd, pszObscure);
		return;
	}
	uid_t i = getuid();
	if (false == change_password_group(szTarget, pszNewPassword)) {
		free(pszNewPassword);
		write_byte(sockfd, ERROR);
		setuid(i);
		return;
	}
	free(pszNewPassword);
	write_byte(sockfd, OK);
	setuid(i);
}

void action_add_user(int sockfd, char *pszLogin)
{
	char szTarget[32];
	char szGecos[128];
	char szDirectory[128];
	char szShell[128];
	int iGroupID;
	int iUserID;
	
	if ((-1 == read_string(sockfd, szTarget, sizeof(szTarget))) ||
		 (-1 == read_string(sockfd, szGecos, sizeof(szGecos))) ||
		 (-1 == read_string(sockfd, szDirectory, sizeof(szDirectory))) ||
		 (-1 == read_string(sockfd, szShell, sizeof(szShell))) ||
		 !read_integer(sockfd, &iUserID) ||
		 !read_integer(sockfd, &iGroupID)) {
		write_byte(sockfd, TIMEOUT);
		return;
	}
	if ('\0' == szDirectory[0])
		strncpy(szDirectory, pConfigFile->GetString("Add", "directory", "/tmp"), sizeof(szDirectory));
	if ('\0' == szShell[0])
		strncpy(szShell, pConfigFile->GetString("Add", "shell", "/dev/null"), sizeof(szShell));
	
	if (strlen(szTarget) < 1 || strlen(szTarget) > 8) {
		write_byte(sockfd, ERROR);
		return;
	}
	for (int q = 0; '\0' != szTarget[q]; q++)
		if (!isalnum(szTarget[q])) {
			write_byte(sockfd, ERROR);
			return;
		}
	
	int j, k = 0;
	for (j = 0; '\0' != szGecos[j]; j++) {
		if (',' == szGecos[j])
			k += 1;
		if (k > 4 ||
			 ';' == szGecos[j] ||
			 '=' == szGecos[j] ||
			 iscntrl(szGecos[j])) {
			write_byte(sockfd, ERROR);
			return;
		}
	}
	if ((j > 0) &&
		 (0 == strcmp(pszLogin, szTarget)) &&
		 (',' != szGecos[strlen(szGecos) - 1])) {
		write_byte(sockfd, ERROR);
		return;
	}
	
	syslog(LOG_USER | LOG_INFO, "user %s adds user %s", pszLogin, szTarget);
	
	if (true == check_target(szTarget)) {
		write_byte(sockfd, INVALID_USER);
		return;
	}
	if (false == allow_add(pszLogin)) {
		write_byte(sockfd, NO_RIGHTS);
		return;
	}
	if (false == check_shell(szShell)) {
		syslog(LOG_USER | LOG_INFO, "bad shell %s for user %s", szShell, pszLogin);
		write_byte(sockfd, BAD_SHELL);
		return;
	}
	
	CPasswdDatabase cPasswdDatabase(true);
	if (NO_ERROR != cPasswdDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading passwd file");
		write_byte(sockfd, ERROR);		
		return;
	}
	CGroupDatabase cGroupDatabase(true);
	if (NO_ERROR != cGroupDatabase.GetError()) {
		syslog(LOG_USER | LOG_INFO, "error reading group file");
		write_byte(sockfd, ERROR);
		return;
	}

	if (-1 == iGroupID) {
		char *pszGroup = pConfigFile->GetString("Add", "group", "nogroup");
		iGroupID = atoi(pszGroup);
		if (false == cGroupDatabase.GetGroup(pszGroup, (gid_t *)&iGroupID))
			if ((char *)NULL == cGroupDatabase.GetName(iGroupID)) {
				syslog(LOG_USER | LOG_INFO, "invalid group for user %s", szTarget);
				write_byte(sockfd, ERROR);
				return;
			}
	}

	if (-1 == iUserID)
		iUserID = cPasswdDatabase.GetMaxUserID() + 1;

	uid_t i = getuid();
	if (false == make_home(szTarget, szDirectory, iUserID, iGroupID)) {
		write_byte(sockfd, ERROR_MAKEDIR);
		setuid(i);
		return;
	}
	if (false == add_user(szTarget, szGecos, szDirectory, szShell, iUserID, iGroupID)) {
		write_byte(sockfd, ERROR);
		setuid(i);
		return;
	}
	write_byte(sockfd, OK);
	setuid(i);
}

void action_getkey(int sockfd)
{
	syslog(LOG_USER | LOG_INFO, "sending public key");

	char *pszPublicKey = pConfigFile->GetString("Security", "public_key", "/etc/passwdd.pbk");
	FILE *fpPublicKey = fopen(pszPublicKey, "rb");
	if (NULL == fpPublicKey) {
		write_byte(sockfd, ERROR);
		return;
	}
	if (0 != fseek(fpPublicKey, 0, SEEK_END)) {
		write_byte(sockfd, ERROR);
		fclose(fpPublicKey);
		return;
	}
	long lFileSize = ftell(fpPublicKey);
	if (-1 == lFileSize)	{
		write_byte(sockfd, ERROR);
		fclose(fpPublicKey);
		return;
	}
	rewind(fpPublicKey);
	void *pvPublicKey = new char[lFileSize];
	if (NULL == fpPublicKey) {
		write_byte(sockfd, ERROR);
		fclose(fpPublicKey);
		return;
	}
	if (lFileSize != (long)fread(pvPublicKey, sizeof(char), lFileSize, fpPublicKey)) {
		write_byte(sockfd, ERROR);
		delete pvPublicKey;
		fclose(fpPublicKey);
		return;
	}
	write_byte(sockfd, OK);
	if (lFileSize != write_data(sockfd, pvPublicKey, lFileSize)) {
		delete pvPublicKey;
		fclose(fpPublicKey);
		return;
	}
	delete pvPublicKey;
	fclose(fpPublicKey);
}

void action_change_password(int sockfd, char *pszLogin, char *pszPassword)
{
	char szTarget[32];
	char *pszNewPassword = NULL;
	int iMin, iMax, iWarn, iInact;
				
	if ((-1 == read_string(sockfd, szTarget, sizeof(szTarget))) ||
		 (NULL == (pszNewPassword = read_password(sockfd))) ||
		 (false == read_integer(sockfd, &iMin)) ||
		 (false == read_integer(sockfd, &iMax)) ||
		 (false == read_integer(sockfd, &iWarn)) ||
		 (false == read_integer(sockfd, &iInact))) {
		write_byte(sockfd, TIMEOUT);
		return;
	}
				
	syslog(LOG_USER | LOG_INFO, "user %s begins changing password of %s", pszLogin, szTarget);
	if (false == check_target(szTarget)) {
		free(pszNewPassword);
		write_byte(sockfd, INVALID_USER);
		return;
	}
	if (false == allow_change(pszLogin, szTarget)) {
		free(pszNewPassword);
		write_byte(sockfd, NO_RIGHTS);
		return;
	}
	char *pszObscure = obscure(pszPassword, pszNewPassword, 0 == strcmp(pszLogin, szTarget));
	if (NULL != pszObscure) {
		free(pszNewPassword);
		write_byte(sockfd, WEAK_PASSWORD);
		write_string(sockfd, pszObscure);
		return;
	}
	uid_t i = getuid();
	if (false == change_password(szTarget, pszNewPassword, iMin, iMax,	iWarn, iInact)) {
		free(pszNewPassword);
		write_byte(sockfd, ERROR);
		setuid(i);
		return;
	}
	free(pszNewPassword);
	write_byte(sockfd, OK);
	setuid(i);
}

void action_add_group(int sockfd, char *pszLogin)
{
	char szTarget[32];
	char szMembers[128];
	char szAdministrators[128];

	if ((-1 == read_string(sockfd, szTarget, sizeof(szTarget))) ||
		 (-1 == read_string(sockfd, szMembers, sizeof(szMembers))) ||
		 (-1 == read_string(sockfd, szAdministrators, sizeof(szAdministrators)))) {
		write_byte(sockfd, TIMEOUT);
		return;
	}

	if (strlen(szTarget) < 1 || strlen(szTarget) > 8) {
		write_byte(sockfd, ERROR);
		return;
	}
	for (int q = 0; '\0' != szTarget[q]; q++)
		if (!isalnum(szTarget[q])) {
			write_byte(sockfd, ERROR);
			return;
		}

	syslog(LOG_USER | LOG_INFO, "user %s tries to add group %s", pszLogin, szTarget);

	if (true == check_target_group(szTarget)) {
		write_byte(sockfd, INVALID_USER);
		return;
	}
	if (false == allow_add(pszLogin)) {
		write_byte(sockfd, NO_RIGHTS);
		return;
	}
	uid_t i = getuid();
	if (false == add_group(szTarget, szMembers, szAdministrators)) {
		write_byte(sockfd, ERROR);
		setuid(i);
		return;
	}
	write_byte(sockfd, OK);
	setuid(i);
}

void action_delete_user(int sockfd, char *pszLogin)
{
	char szTarget[32];

	if (-1 == read_string(sockfd, szTarget, sizeof(szTarget))) {
		write_byte(sockfd, TIMEOUT);
		return;
	}

	syslog(LOG_USER | LOG_INFO, "user %s tries to delete user %s", pszLogin, szTarget);

	if (false == check_target(szTarget)) {
		write_byte(sockfd, INVALID_USER);
		return;
	}
	if (false == allow_change(pszLogin, szTarget)) {
		write_byte(sockfd, NO_RIGHTS);
		return;
	}
	uid_t i = getuid();
	if (false == delete_user(szTarget)) {
		write_byte(sockfd, ERROR);
		setuid(i);
		return;
	}
	write_byte(sockfd, OK);
	setuid(i);
}

void action_chfn(int sockfd, char *pszLogin)
{
	char szTarget[32];
	char szFullName[64];
	char szRoomNumber[64];
	char szWorkPhone[64];
	char szHomePhone[64];
	char szOther[64];				

	if ((-1 == read_string(sockfd, szTarget, sizeof(szTarget))) ||
		 (-1 == read_string(sockfd, szFullName, sizeof(szFullName))) ||
		 (-1 == read_string(sockfd, szRoomNumber, sizeof(szRoomNumber))) ||
		 (-1 == read_string(sockfd, szWorkPhone, sizeof(szWorkPhone))) ||
		 (-1 == read_string(sockfd, szHomePhone, sizeof(szHomePhone))) ||
		 (-1 == read_string(sockfd, szOther, sizeof(szOther)))) {
		write_byte(sockfd, TIMEOUT);
		return;
	}

	if (!check_gecos_field(szFullName) ||
		 !check_gecos_field(szRoomNumber) ||
		 !check_gecos_field(szWorkPhone) ||
		 !check_gecos_field(szHomePhone) ||
		 !check_gecos_field(szOther)) {
		syslog(LOG_USER | LOG_INFO, "invalid gecos for user %s by user %s", pszLogin, szTarget);
		write_byte(sockfd, ERROR);
		return;
	}

	if ((0 == strcmp(pszLogin, szTarget)) && ('\0' != szOther[0])) {
		syslog(LOG_USER | LOG_INFO, "user %s wants to changes his own other fiel", pszLogin);					
		write_byte(sockfd, NO_RIGHTS);
		return;
	}

	syslog(LOG_USER | LOG_INFO, "user %s changes gecos for user %s", pszLogin, szTarget);

	if (false == check_target(szTarget)) {
		write_byte(sockfd, INVALID_USER);
		return;
	}
	if (false == allow_change(pszLogin, szTarget)) {
		write_byte(sockfd, NO_RIGHTS);
		return;
	}
	uid_t i = getuid();
	if (false == change_gecos(szTarget,
									  szFullName,
									  szRoomNumber,
									  szWorkPhone,
									  szHomePhone,
									  szOther)) {
		write_byte(sockfd, ERROR);
		setuid(i);
		return;
	}
	write_byte(sockfd, OK);
	setuid(i);
}

void action_get_group(int sockfd, char *pszLogin)
{
	char szTarget[32];

	if (-1 == read_string(sockfd, szTarget, sizeof(szTarget))) {
		write_byte(sockfd, TIMEOUT);
		return;
	}

	syslog(LOG_USER | LOG_INFO, "user %s tries to get group name of user %s", pszLogin, szTarget);

	if (false == check_target(szTarget)) {
		write_byte(sockfd, INVALID_USER);
		return;
	}
	if (false == allow_change(pszLogin, szTarget)) {
		write_byte(sockfd, NO_RIGHTS);
		return;
	}
	char *pszGroup = get_group(szTarget);
	write_string(sockfd, pszGroup);
	if ('\0' == *pszGroup) {
		write_byte(sockfd, ERROR);
		return;
	}
	write_byte(sockfd, OK);
}

bool check_shell(char *pszShell)
{
	bool fgResult = false;
	setusershell();
	char *pszUserShell;
	while (NULL != (pszUserShell = getusershell()))
		if (0 == strcmp(pszShell, pszUserShell))
			fgResult = true;
	endusershell();
	return fgResult;
}

bool make_home(char *pszTarget, char *pszDirectory, int iUserID, int iGroupID)
{
	if (0 != setuid(0)) {
		syslog(LOG_USER | LOG_INFO, "passwdd needs root privileges to add users; setuid error: %s", strerror(errno));
		return false;
	}
	
	char szNew[256];	
	if (strstr(pszDirectory, "$USER") != (char *)NULL) {
		strncpy(szNew, pszDirectory, sizeof(szNew));
		strcpy(strstr(szNew, "$USER"), pszTarget);
		pszDirectory = szNew;
	}
	
	if (0 != mkdir(pszDirectory, 0755)) {
		syslog(LOG_USER | LOG_INFO, "error creating user directory %s, %s", pszDirectory, strerror(errno));	
		return false;
	}
	if (0 != chown(pszDirectory, iUserID, iGroupID)) {
		syslog(LOG_USER | LOG_INFO, "error executing chown for user directory %s, %s", pszDirectory, strerror(errno));	
		return false;
	}
	
	return true;
}