/*
 * Copyright 1999, Alexander Feldman <alex@varna.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Alexander Feldman nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ALEXANDER FELDMAN AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL ALEXANDER FELDMAN OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <ctype.h>

#include "globals.h"
#include "sockio.h"
#include "conf.hpp"
#include "rsa.hpp"
#include "key.hpp"

#define SERV_TCP_PORT 1099
#define CONF_FILE		 "/etc/passwdc.conf"

#define strzero(p) memset(p, 0, strlen(p))
#define memzero(p) memset(p, 0, sizeof(p))

int main(int argc, char **argv)
{
	static char szServers[512];
	
	static char szPassword[128];
	
	struct sockaddr_in client;
	struct hostent *pHost;
	
	char *pszLogin = (char *)NULL;
	char *pszTarget = (char *)NULL;
	char *pszServer;
	char *pszPassword;
	char *pszFullName = (char *)NULL;
	char *pszRoomNumber = (char *)NULL;
	char *pszWorkPhone = (char *)NULL;
	char *pszHomePhone = (char *)NULL;
	char *pszOther = (char *)NULL;
	
	int sockfd;
	
	int iServerPort;
	
	int iResult = 0;
	
	int iParams = 0;
	
	char *pszConfig = getenv("PASSWDCCONF");
	bool fgError = false;

	if (argc > 1)
		for (int i = 1; i < argc; i++)
			if ('-' == argv[i][0]) {				// We have option
				switch (argv[i][1]) {
					case 'c':							// We have other config file
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszConfig = &argv[i][2];
						} else {
							pszConfig = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszConfig) {
								fprintf(stderr, "Invalid configuration file specified...\n");
								fgError = true;
							}
							i += 1;
						}
					break;
					case 'f':
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszFullName = &argv[i][2];
						} else {
							pszFullName = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszFullName) {
								fprintf(stderr, "Invalid full name specified...\n");
								fgError = true;
							}
							i += 1;
						}
					break;
					case 'r':
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszRoomNumber = &argv[i][2];
						} else {
							pszRoomNumber = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszRoomNumber) {
								fprintf(stderr, "Invalid room number specified...\n");
								fgError = true;
							}
							i += 1;
						}
					break;
					case 'w':
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszWorkPhone = &argv[i][2];
						} else {
							pszWorkPhone = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszWorkPhone) {
								fprintf(stderr, "Invalid work phone specified...\n");
								fgError = true;
							}
							i += 1;
						}
					break;
					case 'h':
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszHomePhone = &argv[i][2];
						} else {
							pszHomePhone = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszHomePhone) {
								fprintf(stderr, "Invalid home phone specified...\n");
								fgError = true;
							}
							i += 1;
						}
					break;
					case 'o':
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszOther = &argv[i][2];
						} else {
							pszOther = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszOther) {
								fprintf(stderr, "Invalid additional info specified...\n");
								fgError = true;
							}
							i += 1;
						}
					break;
					default:
						fprintf(stderr, "Invalid option: %s...\n", argv[i]);
						fgError = true;
					break;
				}
			} else {
				if (0 == iParams)
					pszTarget = argv[i];
				else if (1 == iParams)
					pszLogin = argv[i];
				else
					fgError = true;
				iParams += 1;
			}

	if ((argc < 2) || (argc > 15) || (true == fgError)) {
		printf("Usage: userchfn [<target>] [<user>] [-c <config file>]\n");
		printf("       [-f <full name>] [-r <room number>] [-w <work phone>]\n");
		printf("       [-h <home phone>] [-o <other>]\n");
		printf("Where: <target>    is the user whose gecos you want to change;\n");
		printf("       <login>     is your login if you are master user for the given target.\n");
		printf("Options: -c <config file>\n");
		printf("                   use alternate config file instead the default\n");
		printf("	                  '/etc/passwdc.conf' system wide config file;\n");
		printf("         -f <full name>\n");
		printf("                   user fullname\n");
		printf("         -r <room number>\n");
		printf("                   office number\n");
		printf("         -w <work phone>\n");
		printf("                   office extension\n");
		printf("         -h <home phone>\n");
		printf("                   home phone\n");
		printf("         -o <other>\n");
		printf("                   additional info\n");
		printf("Written by: <Alexander Feldman> alex@varna.net\n\n");
		return 1;
	}
	
	if (NULL == pszConfig)
		pszConfig = CONF_FILE;
	
	char *pszMessage = NULL;	
	CConfigurationFile *pConfigFile = new CConfigurationFile(pszConfig, &pszMessage);
	if (NULL == pConfigFile) {
		fprintf(stderr, "Error allocating memory...\n");
		return 1;
	}
	if (NULL != pszMessage) {
		fprintf(stderr, "%s\n", pszMessage);
		delete pConfigFile;
		return 1;
	}
	
	iServerPort = pConfigFile->GetInteger("Global", "port", SERV_TCP_PORT);
	memzero(szServers);
	strncpy(szServers, pConfigFile->GetString("Global", "targets", "localhost"), sizeof(szServers) - 1);
	
	openlog(argv[0], LOG_PID, LOG_USER);
	
	if (NULL == pszLogin)
//		pszLogin = getlogin();				// This returns NULL in some glibc
		pszLogin = getenv("LOGNAME");		// Thanks to Mark Siegler <tech@netalia.com>
	
	if (NULL == pszLogin) {
		fprintf(stderr, "Error getting login from the environment...\n");
		return 1;
	}	
	if (NULL == pszTarget)
		pszTarget = pszLogin;
	
	if ((NULL != pszOther) && (0 == strcmp(pszTarget, pszLogin))) {
		fprintf(stderr, "Only the superuser has the right to change the additional info...\n");
		delete pConfigFile;
		return 1;
	}
	
	pszPassword = getpass("Password:");
	strcpy(szPassword, pszPassword);
	strzero(pszPassword);

	pszServer = strtok(szServers, ";,");
	do {
		pszServer += strspn(pszServer, " \t");			
		printf("Chfn for %s on %s\n", pszTarget, pszServer);

		char szFileName[256];
		
		char *pszKeyDirectory = pConfigFile->GetString("Security", "key_directory", "/var/spool/passwdc");
		
		strncpy(szFileName, pszKeyDirectory, sizeof(szFileName));
		strncat(szFileName, "/", sizeof(szFileName));
		strncat(szFileName, pszServer, sizeof(szFileName));
		
		if (false == RSALoadPublicKey(szFileName)) {
			printf("Requesting public key from server %s.\n", pszServer);
			get_key(pszServer, iServerPort, pszKeyDirectory);
			RSALoadPublicKey(szFileName);
		}
	
		if ((sockfd = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
			printf("Socket error: %s\n", strerror(errno));
			continue;
		}

		pHost = gethostbyname(pszServer);
		if (NULL == pHost) {
			printf("Resolving error: %s\n", pszServer);
			close(sockfd);
			continue;
		}
		memset(&client, 0, sizeof(client));
		client.sin_family = AF_INET;
		client.sin_addr.s_addr = *((unsigned long int *)(pHost->h_addr));
		client.sin_port = htons(iServerPort);
		if (connect(sockfd, (struct sockaddr *)&client, sizeof(client)) < 0) {
			printf("Connect error: %s\n", strerror(errno));
			close(sockfd);
			continue;
		}

		unsigned char c = '\0';
		read_byte(sockfd, &c);
		if (WELCOME != c) {
			printf("Invalid response from server.\n");
			close(sockfd);
			continue;
		} 
		if (((int)strlen(pszLogin) + 1 != write_string(sockfd, pszLogin)) ||
			 (false == write_password(sockfd, szPassword))) {
			printf("Write error.\n");
			close(sockfd);
			continue;
		} 
		read_byte(sockfd, &c);			
		switch (c) {
			case GO_AHEAD:
				if (NULL == pszFullName)
			  		pszFullName = "";
				if (NULL == pszRoomNumber)
			  		pszRoomNumber = "";
				if (NULL == pszWorkPhone)
			  		pszWorkPhone = "";
				if (NULL == pszHomePhone)
			  		pszHomePhone = "";
				if (NULL == pszOther)
			  		pszOther = "";
				write_byte(sockfd, ACTION_CHFN);
				if (((int)strlen(pszTarget) + 1 != write_string(sockfd, pszTarget)) ||
					 ((int)strlen(pszFullName) + 1 != write_string(sockfd, pszFullName)) ||
					 ((int)strlen(pszRoomNumber) + 1 != write_string(sockfd, pszRoomNumber)) ||
					 ((int)strlen(pszWorkPhone) + 1 != write_string(sockfd, pszWorkPhone)) ||
					 ((int)strlen(pszHomePhone) + 1 != write_string(sockfd, pszHomePhone)) ||
					 ((int)strlen(pszOther) + 1 != write_string(sockfd, pszOther))) {
					printf("Write error.\n");
					close(sockfd);
					continue;
				}
				read_byte(sockfd, &c);
				switch (c) {
					case OK:
						printf("Operation successfully completed.\n");
					break;
					case INVALID_USER:
						printf("Invalid target specified.\n");
					break;
					case ERROR:
						printf("Invalid gecos. Server returned error.\n");
					break;
					case NO_RIGHTS:
						printf("Access denied.\n");
					break;
					default:
						printf("Invalid response from server.\n");
				}
			break;
			case LOGIN_FAILED:
				printf("Incorrect login or password for %s on %s.\n", pszLogin, pszServer);				
			break;
			default:
				printf("Invalid response from server.\n");
		}
		close(sockfd);
	} while (NULL != (pszServer = strtok(NULL, ";,")));
	
	closelog();

	strzero(szPassword);

	delete pConfigFile;
	
	return iResult;
}